/*****************************************************************
 * sigknock.c
 *
 * Signature knocking client
 *
 * Signs a challenge obtained from a sigknock-enabled server
 * and sends it back.
 *
 * Author: Thomas Wana <thomas@wana.at>
 *
 * $Id: sigknock.c 143 2005-01-12 10:31:08Z greuff $
 *
 *****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/select.h>
#include <netdb.h>
#include <string.h>
#include <arpa/inet.h>
#include <time.h>
#include <linux/netfilter_ipv4/ipt_sigknock.h>
#include <errno.h>
#include "../config.h"

#define SYNCOOKIE_HASH_LEN 20

int sock=0;
char syncookie_hash[SYNCOOKIE_HASH_LEN];

/************************************************************************/

// A routine that prints a buffer in 'hexdump -C'-style via printf
void hexdump(unsigned char *buffer, long len)
{
   unsigned long pos=0;

   while(len>0)
   {
      int i;
      printf("%08x  ",(unsigned int)pos);
      for(i=0;i<16 && i<len;i++)
      {
         printf("%02x ",buffer[pos+i]);
         if(i==7 || i==15)
            printf(" ");
      }
      printf("|");
      for(i=0;i<16 && i<len;i++)
      {
         unsigned char c=buffer[pos+i];
         if( (c>='A' && c<='Z') || (c>='a' && c<='z') || (c>='0' && c<='9') ||
             c==':' || c=='!' || c=='?' || c=='=' || c==';' || c=='<' ||
             c=='>' || c=='/' || c=='\\' || c=='&')
         {
            printf("%c",c);
         }
         else
         {
            printf(".");
         }
      }
      printf("|\n");
      len-=16;
      pos+=16;
   }
}

/************************************************************************/

/*
 * Just fire out the payload.
 * Return: -1 for error in syscall (errno set)
 *          0 on success
 */
int send_no_wait(int sock, char *buf, int len)
{
	if(send(sock, buf, len, 0)<0)
	{
		return -1;
	}
	return 0;
}

/************************************************************************/

// Return: -1 for error in syscall (errno set)
//         -2 for timeout
//         0 on success, data ready to be picked up from socket
int send_and_wait_for_response_with_retries(int sock, char *buf, int len)
{
   int i, res;
   struct timeval tv;
   fd_set rdset;

   for(i=0;i<3;i++)
   {
      if(send(sock, buf, len, 0)<0)
      {
         return -1;
      }
      
      FD_ZERO(&rdset);
      FD_SET(sock, &rdset);
      tv.tv_sec=5;
      tv.tv_usec=0;
      res=select(sock+1, &rdset, NULL, NULL, &tv);
      if(res==1) break;
      if(res<0)
         return -1;

      printf("timeout, retrying (%d/3) ... ",i+1);
      fflush(stdout);
   }
   if(i==3)
   {
      printf("timed out, aborting\n");
      errno=ETIME;
      return -2;
   }

   return 0;
}

// Requests a challenge from the server.
// Returns a pointer to the freshly malloc'd challenge and writes the
// length of the challenge to *len.
char *request_challenge(unsigned int *outlen)
{
   char *rval=NULL;
   char buf[5000];
   int len;
   
   // obtain the communication syncookie
   memset(buf,0,sizeof(buf));
   buf[0]=COOKIE_REQUEST;
   if(send_and_wait_for_response_with_retries(sock, buf, 1)<0)
   {
      perror("send_and_wait_for_response_with_retries");
      return NULL;
   }
   
   if((len=recv(sock, buf, sizeof(buf), 0))<0)
   {
      perror("recv");
      return NULL;
   }
   if(len!=1+SYNCOOKIE_HASH_LEN)
   {
      printf("wrong length of return packet\n");
      return NULL;
   }
   if(buf[0]!=COOKIE_RESPONSE)
   {
      printf("protocol synchronization error\n");
      return NULL;
   }
   memcpy(syncookie_hash, buf+1, SYNCOOKIE_HASH_LEN);

   // request the challenge
   buf[0]=CHALLENGE_REQUEST;
   if(send_and_wait_for_response_with_retries(sock, buf, 1+SYNCOOKIE_HASH_LEN)<0)
   {
      perror("send_and_wait_for_response_with_retries");
      return NULL;
   }

   // read the response
   if((len=recv(sock, buf, sizeof(buf), 0))<0)
   {
      perror("recv");
      return NULL;
   }
   if(len!=1+SYNCOOKIE_HASH_LEN+CHALLENGE_LEN)
   {
      printf("wrong length of return packet (2)\n");
      return NULL;
   }
   if(buf[0]!=CHALLENGE_RESPONSE)
   {
      printf("protocol synchronization error (2)\n");
      return NULL;
   }

   *outlen=len-1-SYNCOOKIE_HASH_LEN;
   rval=(char *)malloc(*outlen);
   if(rval==NULL)
   {
      printf("out of memory\n");
      return NULL;
   }

   memcpy(rval, buf+1+SYNCOOKIE_HASH_LEN, *outlen);

   return rval;
}

/************************************************************************/

// signes the challenge using gpg. 
// Returns a pointer to the binary signature returned by
// gpg (new malloc'd space)
char *sign_challenge(char *uid, char *challenge, unsigned int len, unsigned int *outlen)
{
   char tmpname[]="/tmp/pktmpXXXXXX";
   char cmd[200];
   int fd;
   FILE *pp;
   char *rval;

   rval=(char *)malloc(65000);
   if(rval<0)
   {
      printf("out of memory\n");
      return NULL;
   }

   // write challenge to file
   fd=mkstemp(tmpname);
   if(fd<0)
   {
      perror("mkstemp");
      free(rval);
      return NULL;
   }
#ifdef DEBUG
   printf("using tmp file: %s\n",tmpname);
#endif
   if(write(fd, challenge, len)<0)
   {
      perror("write");
      free(rval);
      unlink(tmpname);
      return NULL;
   }
   close(fd);

   // sign the file (detached binary signature)
   snprintf(cmd, sizeof(cmd), "%s -u %s --output - --detach-sign %s",GPG_PATH,uid,tmpname);

   printf("----- gpg output from here on -----\n");

   fflush(stdin);
   fflush(stdout);
   fflush(stderr);
   pp=popen(cmd, "r");
   if(pp==NULL)
   {
      perror("popen");
      free(rval);
      unlink(tmpname);
      return NULL;
   }
   *outlen=fread(rval, 1, 65000, pp);
   if(ferror(pp))
   {
      printf("error reading pipe\n");
      free(rval);
      unlink(tmpname);
      return NULL;
   }
   pclose(pp);

   printf("----- gpg output finished -----\n");

   unlink(tmpname);

   return rval;
}

/************************************************************************/

// sends the signature back to the server
// returns 0 on success, <0 otherwise
int open_port(char *signature, unsigned int signlen)
{
   char *payload;
   
   payload=(char *)malloc(signlen+SYNCOOKIE_HASH_LEN+1);
   if(payload==NULL)
   {
      printf("out of memory\n");
      return -1;
   }
   payload[0]=SIGNATURE_REQUEST;
   memcpy(payload+1, syncookie_hash, SYNCOOKIE_HASH_LEN);
   memcpy(payload+1+SYNCOOKIE_HASH_LEN, signature, signlen);
   
   // send it out
   if(send_no_wait(sock, payload, SYNCOOKIE_HASH_LEN+1+signlen)<0)
   {
      perror("send_and_wait_for_response_with_retries");
      free(payload);
      return -1;
   }

   free(payload);

   return 0;
}

/************************************************************************/

void inline print_banner()
{
   printf("SigKnock %s: Client for Linux, BSD, MacOS X\n\n",SIGKNOCK_VERSION);
}

/************************************************************************/

void print_help()
{
   printf("Error - insufficient parameters\n\n");
   printf("Usage: ./sigknock [options]\n\n");
   printf("Options:\n"
          "\t-h:\t\tHostname of the server to knock on\n"
          "\t-k:\t\tKeyport to knock on\n"
          "\t-u:\t\tUser-ID to use when invoking gpg for signature creation\n\n");
   printf("Example:\n"
          "./sigknock -h top.secret.nsa.gov -k 3000 -u george.w@whitehouse.gov\n\n");
}

/************************************************************************/

int main(int argc, char **argv)
{
   struct in_addr host;
   struct sockaddr_in sin;
   unsigned int keyport=0;
   char uid[50];
   char *challenge, *signature;
   unsigned int opt, len, signlen;
   unsigned int param_seen=0;

   print_banner();
   
   while((opt=getopt(argc, argv, "h:k:u:"))!=-1)
   {
      switch(opt)
      {
         case 'h':
            {
               struct hostent *he;
               he=gethostbyname(optarg);
               if(!he)
               {
                  herror("gethostbyname");
                  return -1;
               }
               host.s_addr=*(in_addr_t *)he->h_addr_list[0];

               param_seen|=1;
            }
            break;
         case 'k':
            keyport=(unsigned int)strtol(optarg, NULL, 10);
            param_seen|=2;
            break;
         case 'u':
            snprintf(uid,sizeof(uid),"%s",optarg);
            param_seen|=4;
            break;
         case '?':
            return -1;
            break;
         default:
            fprintf(stderr,"getopt error\n");
            return -1;
      }
   }
   if(param_seen!=0x7)
   {
      // not all parameters supplied
      print_help();
      return -1;
   }

   printf("Starting signature knock on host '%s', keyport %u with UID '%s' ...\n",
         inet_ntoa(host), keyport, uid);
   
   // open UDP socket
   if((sock=socket(PF_INET, SOCK_DGRAM, 0))<0)
   {
      perror("socket");
      return -3;
   }

   // connect socket
   sin.sin_addr=host;
   sin.sin_port=htons(keyport);
   sin.sin_family=AF_INET;
   if(connect(sock, (struct sockaddr *)&sin, sizeof(sin))<0)
   {
      perror("connect");
      return -3;
   }

   // get the challenge
   printf("Requesting challenge ... ");
   fflush(stdout);
   challenge=request_challenge(&len);
   if(challenge==NULL)
   {
      printf("error requesting challenge.\n");
      return -2;
   }
   printf("success.\n");
#ifdef DEBUG
   hexdump(challenge, len);
#endif

#ifdef DEBUG
   printf("got syncookie: \n");
   hexdump(syncookie_hash, SYNCOOKIE_HASH_LEN);
#endif

   // use gnupg to sign the challenge
   printf("Signing challenge ...\n");
   fflush(stdout);
   signature=sign_challenge(uid, challenge, len, &signlen);
#ifdef DEBUG
   printf("successfully signed challenge:\n");
   hexdump(signature, signlen);
#endif
   free(challenge);

   // send the signature back to the server
   printf("Executing signature knock ... ");
   fflush(stdout);
   if(open_port(signature, signlen)<0)
   {
      printf("error sending the signature\n");
   }
   else
      printf("success.\n");

   free(signature);

   return 0;
}
