using System;
using System.Net;
using System.Net.Sockets;
using System.ComponentModel;
using System.Collections;

namespace SigKnockOpener
{
	/// <summary>
	/// Summary description for SigKnockOpener.
	/// </summary>
	public class SigKnockOpener
	{
		public SigKnockOpener()
		{
			
		}

		/// <summary>
		/// Some Events that are fired during the sigknock procedure
		/// </summary>
		public event EventHandler GettingSyncookie;
		public event EventHandler GotSyncookie;
		public event EventHandler GettingChallenge;
		public event EventHandler GotChallenge;
		public event EventHandler GettingSignature;
		public event EventHandler GotSignature;
		public event EventHandler StartingSigKnock;
		public event EventHandler SigKnockDone;

		/// <summary>
		/// This event is fired when a sending/receiving retry is attempted.
		/// </summary>
		public event EventHandler Retry;

		public byte[] SendWithRetryAndReceive(Socket s, byte[] buffer, int count)
		{
			byte[] recvbuf=new byte[65535];
			byte[] rval=null;
			bool done=false;
			int i;
			
			// select for readability from the socket
			for(i=0;i<3 && done==false;i++)
			{
				// send data
				s.Send(buffer, count, SocketFlags.None);

				// check if there is something to read on the socket
				ArrayList fdread=new ArrayList();
				fdread.Add(s);
				Socket.Select(fdread, null, null, 2*1000*1000);
				if(fdread.Count==1)
				{
					done=true;
				}
				/*if(s.Poll(2*1000*1000, SelectMode.SelectRead)==true)
				{
					// there is, finish the loop
					done=true;
				}*/
				if(done==false)
				{
					// retrying
					if(Retry!=null)
						Retry(this, new EventArgs());
				}
			}
			if(done==true)
			{
				count=s.Receive(recvbuf);
				rval=new byte[count];
				Array.Copy(recvbuf, 0, rval, 0, count);
			}
			else
			{
				throw new Exception("Timeout while waiting for a reply from the server");
			}
			return rval;
		}

		/// <summary>
		/// Executes the SigKnock.
		/// </summary>
		/// <param name="UID">User-ID for the Signhelper</param>
		/// <param name="Host">Host to knock against</param>
		/// <param name="Keyport">Keyport on the host</param>
		/// <param name="helper">Reference to the Signature helper class (OS-Independence)</param>
		/// <returns></returns>
		public bool knock(String UID, String Host, String Keyport, ref SigKnockSignHelper helper)
		{
			Socket sock;
			byte[] buffer=new byte[1];
			IPEndPoint endpoint=null;

			// create the socket and establish connection
			sock=new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
			endpoint=new IPEndPoint(Dns.Resolve(Host).AddressList[0], int.Parse(Keyport));
			sock.Connect(endpoint);

			// Request a syncookie
			buffer[0]=0;
			if(GettingSyncookie!=null)
				GettingSyncookie(this, new EventArgs());
			buffer=this.SendWithRetryAndReceive(sock, buffer, 1);
			if(buffer[0]!=1)
			{
				throw new Exception("Protocol error (syncookie request)");
			}
			if(GotSyncookie!=null)
				GotSyncookie(this, new EventArgs());

			// copy away the syncookie
			byte[] syncookie=new byte[buffer.Length-1];
			Array.Copy(buffer, 1, syncookie, 0, buffer.Length-1);

			// use the syncookie response to form the challenge request
			buffer[0]=2;
			if(GettingChallenge!=null)
				GettingChallenge(this, new EventArgs());
			buffer=this.SendWithRetryAndReceive(sock, buffer, buffer.Length);
			if(buffer[0]!=3)
			{
				throw new Exception("Protocol error (challenge request)");
			}
			if(GotChallenge!=null)
				GotChallenge(this, new EventArgs());

			// get the challenge
			byte[] challenge=new byte[buffer.Length-syncookie.Length-1];
			Array.Copy(buffer, 1+syncookie.Length, challenge, 0, buffer.Length-syncookie.Length-1);

			// now execute the helper to get the signature
			if(GettingSignature!=null)
				GettingSignature(this, new EventArgs());
			byte[] signature=helper.sign(challenge, UID);
			if(GotSignature!=null)
				GotSignature(this, new EventArgs());

			// and execute the signature request
			byte[] sigrequest=new byte[1+syncookie.Length+signature.Length];
			sigrequest[0]=4;
			Array.Copy(syncookie, 0, sigrequest, 1, syncookie.Length);
			Array.Copy(signature, 0, sigrequest, 1+syncookie.Length, signature.Length);
			if(StartingSigKnock!=null)
				StartingSigKnock(this, new EventArgs());
			sock.Send(sigrequest, sigrequest.Length, SocketFlags.None);

			// TODO get sigknock response
			if(SigKnockDone!=null)
				SigKnockDone(this, new EventArgs());
			
			return true;
		}
	}
}
