/************************************************************
 * libipt_sigknock.c
 *
 * Userspace-iptables-interface to sigknock.
 *
 * Author: Thomas Wana <thomas@wana.at>
 *
 * $Id: libipt_sigknock.c 114 2004-08-27 10:53:42Z greuff $
 *
 ************************************************************/


#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#if defined(__GLIBC__) && __GLIBC__ == 2
#include <net/ethernet.h>
#else
#include <linux/if_ether.h>
#endif
#define __user
#include <iptables.h>
#include <linux/netfilter_ipv4/ipt_sigknock.h>
#include <time.h>
#include <unistd.h>

#include "../config.h"

int cmdlineflags;

/* Function which prints out a usage message. */
static void help_match(void)
{
	printf(
"Signature knocking v%s options:\n"
" --uid User-ID\n"
"			   User-ID of the signature to check against. You can use this several times in a single rule.\n"
" --timeout Timeout[smhd]\n"
"			   Timeout (in seconds, minutes, hours or days) after the port closes again\n"
" --close-after-syn\n"
"			   Close the match after the first TCP-SYN came through\n"
"timeout and close-after-syn are exclusive.\n"
"Note: you *must* have a SIGKNOCK-KEYPORT target in your fw-rules; see\n"
"the sigknock docs for details.\n"
"\n", SIGKNOCK_VERSION);
}

static struct option opts[] = {
	{ "timeout", 1, 0, '1' },
	{ "uid", 1, 0, '2' },
	{ "close-after-syn", 0, 0, '3' },
	{0}
};

/* Initialize the match. */
static void init_match(struct ipt_entry_match *m, unsigned int *nfcache)
{
   struct ipt_sigknock_info *pkinfo=(struct ipt_sigknock_info *)m->data;
   int i;
	/* Can't cache this */
	*nfcache |= NFC_UNKNOWN;

   cmdlineflags=0;

   memset(pkinfo, 0, sizeof(struct ipt_sigknock_info));

   pkinfo->close_after_syn=0;
   pkinfo->timeout=-1;
   pkinfo->timeout_char=' ';
   pkinfo->uid_count=0;

   // create the match-ID (see ipt_sigknock.h for details)
   for(i=0;i<sizeof(pkinfo->id);i++)
      pkinfo->id[i]=rand()%256;
}

/* Function which parses command options; returns true if it
   ate an option */
static int parse_match(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_sigknock_info *pkinfo = (struct ipt_sigknock_info *)(*match)->data;
   char *ch;

   // TODO: better sanity checks
	switch (c) {
	case '1': // timeout
      pkinfo->timeout=(int)strtol(argv[optind-1],&ch,10);
      pkinfo->timeout_char=*ch;
      if(*ch!='s' && *ch!='m' && *ch!='h' && *ch!='d')
      {
         // parameter problem
         return 0;
      }
      cmdlineflags|=2;
		break;
	case '2': // uid
      if(pkinfo->uid_count+1>=sizeof(pkinfo->uid)/sizeof(pkinfo->uid[0]))
      {
         // already used all UID slots
         cmdlineflags=-1;
         return 0;
      }
      snprintf(pkinfo->uid[pkinfo->uid_count],sizeof(pkinfo->uid[pkinfo->uid_count]),
            "%s",argv[optind-1]);
      pkinfo->uid_count++;
      cmdlineflags|=1;
		break;
   case '3': // close-after-syn
      pkinfo->close_after_syn=1;
      cmdlineflags|=4;
      break;
	default:
		return 0;
	}

	return 1;
}

/* Final check; all mandatory parameters have to be given */
static void final_check_match(unsigned int flags)
{
   if(cmdlineflags==-1)
      exit_error(PARAMETER_PROBLEM, "You can't use that many UIDs in one rule; sorry.");
   if ( (cmdlineflags&2) && (cmdlineflags&4) )
      // TODO auslaenderenglisch
      exit_error(PARAMETER_PROBLEM, "Error: you can't use --timeout and --close-after-syn together");
	if (cmdlineflags!=0x3 && cmdlineflags!=0x5)
		exit_error(PARAMETER_PROBLEM, "Parameters missing / syntax error");
}

/* Prints the matchinfo. */
static void print_match(const struct ipt_ip *ip,
      const struct ipt_entry_match *match,
      int numeric)
{
   struct ipt_sigknock_info *info=(struct ipt_sigknock_info *)match->data;
   int i;

	printf("Signature knock: User-IDs: ");
   for(i=0;i<info->uid_count && i<sizeof(info->uid)/sizeof(info->uid_count);++i)
      printf("'%s', ",info->uid[i]);
   if(info->close_after_syn)
      printf("close after SYN");
   else
      printf("Timeout %u%c",info->timeout,info->timeout_char);
   printf(" ");
}

/* Saves the matchinfo in parseable form to stdout. */
static void save_match(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
   int i;
   struct ipt_sigknock_info *info=(struct ipt_sigknock_info *)match->data;
   for(i=0;i<info->uid_count && i<sizeof(info->uid)/sizeof(info->uid_count);++i)
	   printf("--uid \"%s\" ",info->uid[i]);
   if(info->close_after_syn)
      printf("--close-after-syn ");
   else
      printf("--timeout %u%d ",info->timeout, info->timeout_char);
}

static struct iptables_match sigknock_match = { 
    .next=NULL,
    .name="sigknock",
    .version=NETFILTER_VERSION,
    .size=IPT_ALIGN(sizeof(struct ipt_sigknock_info)),
    .userspacesize=IPT_ALIGN(sizeof(struct ipt_sigknock_info)),
    .help=&help_match,
    .init=&init_match,
    .parse=&parse_match,
    .final_check=&final_check_match,
    .print=&print_match,
    .save=&save_match,
    .extra_opts=opts,
};

void _init(void)
{
	register_match(&sigknock_match);
   srand(time(NULL)*getpid());
}
