/**************************************************************************
 * sigknock.c
 *
 * by Thomas Wana <thomas@wana.at>
 *
 * The main signature knocking module. 
 *
 * $Id: sigknock.c 146 2005-01-12 12:48:13Z greuff $
 *
 **************************************************************************/

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#ifndef LinuxVersionCode
/* Very old kernels don't have an equivalent macro... */
#define LinuxVersionCode(v, p, s) (((v)<<16)+((p)<<8)+(s))
#endif

#include <linux/version.h>
#if LINUX_VERSION_CODE < LinuxVersionCode(2,5,0)
	#define __KERNEL_SYSCALLS__
#endif
#include <linux/config.h>
#ifdef CONFIG_SMP
	#define __SMP__
#endif
#ifndef CONFIG_NETFILTER
	#error You need netfilter built into the kernel for this module to work.
#endif
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/if_ether.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/random.h>
#include <linux/wait.h>
#include <linux/kmod.h>
#include <linux/list.h>
#include <linux/proc_fs.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/signal.h>
#if LINUX_VERSION_CODE < LinuxVersionCode(2,5,0)
	#include <linux/unistd.h>
#else
	#include <linux/syscalls.h>
#endif
#include <net/udp.h>
#include <net/route.h>
#include <net/flow.h>
#include <asm/uaccess.h>
#include <asm/checksum.h>

#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_sigknock.h>
#include "../config.h"
#include "cipher-sha1.h"

#ifndef __i386__
// #warning "This module doesn't work on non-i386 architectures!!"
#endif

#ifdef DEBUG
#define PRINTK printk
#else
#define PRINTK(a...)
#endif

// #define ULTRADEBUG        // define this to include debug code even for
                             // code parts that I am sure to work

#define CALLER_ID_SIGNATURE "GPG_CALLER_ID="
#define CALLER_ID_LEN 10        // number of bytes of the caller id
#define SYNCOOKIE_HASH_SIZE 20  // length of the syncookie hash, 20 bytes is SHA-1
#define KNOCK_THRESHOLD 5       // after how many seconds the next knock is allowed

#define ADDRARGS(addr) (addr)&0xFF,((addr)>>8)&0xFF,((addr)>>16)&0xFF,((addr)>>24)&0xFF

/*
 * It's such a PITA ...
 */
#if LINUX_VERSION_CODE < LinuxVersionCode(2,5,0)
	#define sys_open open
	#define sys_read read
	#define sys_close close
#endif

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Thomas Wana <thomas@wana.at>");
MODULE_DESCRIPTION("iptables signature knocking module");

static int errno;

/*
 * kernel threads synchronization stuff
 */
static DECLARE_WAIT_QUEUE_HEAD(signature_wait_queue);
static int exit_signature_thread_command=0;
static struct semaphore sem_exit_notify;

/*
 * A list containing the relationship between a source address,
 * the issued challenge and the returning signature. Using it,
 * the signature thread can confirm the validity of the signature.
 * Primary key is the source address.
 */
struct auth_list_elem
{
	struct list_head list;

	/*
	 * source address the challenge was issued to.
	 * We only accept responses from the same source that we
	 * issued the challenge to.
	 */
	__u32 saddr;

	/*
	 * the issued challenge
	 */
	unsigned char challenge[CHALLENGE_LEN];

	/*
	 * timestamps we issued the challenge and got the last signature,
	 * so we can implement packet throttling and prevent DoSes 
	 */
	unsigned long challenge_timestamp;
	unsigned long signature_timestamp;

	/*
	 * flag for the signature thread so it knows
	 * whether to check the signature in here
	 */
	int needs_check;

	/*
	 * received signature with length
	 */
	unsigned char signature[4096];
	size_t siglen;

	/*
	 * the callerid - so when gpg reads the challenge and signature
	 * from /proc, the read-handler knows what challenge and signature
	 * to return and where to write the result to
	 */
	unsigned char callerid[CALLER_ID_LEN];

	/*
	 * the result coming from gpg (which writes it in /proc/result)
	 */
	unsigned char result[4096]; // should never be larger than that
	size_t resultlen;
};

static LIST_HEAD(auth_list);
/*
 * spinlock for the manipulation of the list's structure
 */
static spinlock_t auth_list_lock=SPIN_LOCK_UNLOCKED;

/*
 * A list containing all the sigknock-rules that are currently
 * active.
 */
struct info_list_elem
{
	struct list_head list;

	/*
	 * the info coming from iptables
	 */
	struct ipt_sigknock_info pkinfo;

	/*
	 * the master switch: determines if the rule matches or not
	 */
	int authorized;

	/*
	 * The timestamp when the port opened; so we can check if it is
	 * still within the timeout. timediff holds the calculated timeout
	 * (in jiffies) for fast checking in the match function
	 */
	unsigned long timestamp;
	unsigned long timediff;
};

static LIST_HEAD(info_list);
/*
 * spinlock for the manipulation of the list's structure
 */
static spinlock_t info_list_lock=SPIN_LOCK_UNLOCKED;

/*
 * we use the ideas of syncookie-protection to get rid of spoofed
 * source hosts. This is the structure we run our secret function
 * across.
 */
struct syncookie_struct
{
	unsigned long timestamp;
	pid_t sigthread_pid;
	__u32 saddr;
	__u16 sport;
	__u32 daddr;
	__u16 dport;
};

/*
 * Proc-FS interface
 */
static struct proc_dir_entry *proc_dir, *proc_challenge, *proc_signature, *proc_result;

/*
 * PID of the signature thread, stored globally for usage in the
 * syncookies-struct (read-only, except on startup after kernel_thread -
 * no lock needed)
 */
static pid_t sigthread_pid;

/*
 * jiffies shift count (see init for details)
 */
static int jiffies_shift_for_delay;

static void dump_infolist(struct list_head *l);
static void dump_authlist(struct list_head *l);

/*******************************************************************************/

/*
 * compatibility code - 2.6 has dst_output, 2.4 doesn't
 */
static inline int my_dst_output(struct sk_buff *skb)
{
#if LINUX_VERSION_CODE < LinuxVersionCode(2,5,0)
	int err;
	for(;;)
	{
		err=skb->dst->output(skb);
		if(err==0)
			return err;
		if(err!=NET_XMIT_BYPASS)
			return err;
	}
#else
	return dst_output(skb);
#endif
}

/*******************************************************************************/

/*
 * returns 0 if the contents of the two elements are the same,
 * !0 otherwise.
 */
static inline int infocmp(struct ipt_sigknock_info *elem1, 
            struct ipt_sigknock_info *elem2)
{
	int rval=1;
	int i;

	if(elem1->uid_count!=elem2->uid_count)
		return 1;

	/*
	 * compare UIDs
	 */
	for(i=0;i<elem1->uid_count && i<sizeof(elem1->uid)/sizeof(elem1->uid[0]);++i)
		if(strncmp(elem1->uid[i], elem2->uid[i], sizeof(elem2->uid[i])))
			return 1;

	/*
	 * compare the rest
	 */
	if(elem1->timeout==elem2->timeout &&
			elem1->timeout_char==elem2->timeout_char &&
			elem1->close_after_syn==elem2->close_after_syn &&
			memcmp(elem1->id, elem2->id, sizeof(elem2->id))==0)
	{
		rval=0;
	}
	return rval;
}

/*******************************************************************************/

static int get_callerid_from_environment(struct task_struct *task, char id[CALLER_ID_LEN])
{
	int res=-1, len;
	int fd;
	char fname[50], *clrid, *tmppage;
	mm_segment_t oldfs;
	int i;

	/* struct mm_struct *mm = get_task_mm(task);
	   if(mm)
	   {
	   unsigned int len=mm->env_end - mm->env_start;
	   if(len>PAGE_SIZE)
	   len=PAGE_SIZE;

	// allocate page
	 *page=(char *)__get_free_page(GFP_KERNEL);
	 if(*page==NULL)
	 {
	 mmput(mm);
	 PRINTK(KERN_ERR "%s: out of memory\n",__FUNCTION__);
	 return -1;
	 }
	 res=access_process_vm(task, mm->env_start, *page, len, 0);
	 mmput(mm);      
	 *plen=len;
	 } */

	/*
	 * the code above doesn't work; the kernel doesn't export
	 * the necessary functions.
	 * I'm all ears for a better way to get current's
	 * environment.
	 */

	/*
	 * allocate page
	 */
	tmppage=(char *)__get_free_page(GFP_KERNEL);
	if(tmppage==NULL)
	{
		printk(KERN_ERR "%s: out of memory\n",__FUNCTION__);
		return -1;
	}
	/*
	 * read current environment from procfs
	 */
	snprintf(fname, sizeof(fname), "/proc/%d/environ", current->tgid); // SMP safe (sys_getpid)
	oldfs=get_fs();
	set_fs(get_ds());
	fd=sys_open(fname, O_RDONLY, 0);
	set_fs(oldfs);
	if(fd<0)
	{
		printk("%s: error opening %s for reading: %d\n",__FUNCTION__,fname, errno);
		goto fail0;
	}
	memset(tmppage,0,PAGE_SIZE);
	oldfs=get_fs();
	set_fs(get_ds());
	res=sys_read(fd, tmppage, PAGE_SIZE-1);
	set_fs(oldfs);
	if(res<0)
	{
		printk("%s: read failed: %d\n",__FUNCTION__, errno);
		sys_close(fd);
		goto fail0;
	}
	sys_close(fd);

	/*
	 * find caller ID in environment
	 */
	len=res;
	/*
	 * extremely slow and lame memmem, but there is no memmem
	 * in the kernel lib and I am too lazy to write a better one
	 */
	clrid=NULL;
	for(i=0;i<(PAGE_SIZE-CALLER_ID_LEN-strlen(CALLER_ID_SIGNATURE)) && clrid==NULL;i++)
	{
		if(memcmp(tmppage+i, CALLER_ID_SIGNATURE, strlen(CALLER_ID_SIGNATURE))==0)
		{
			clrid=tmppage+i+strlen(CALLER_ID_SIGNATURE);
		}
	}

	if(clrid==NULL)
	{
		printk(KERN_WARNING "%s: unable to find caller id signature in environment\n",__FUNCTION__);
		goto fail0;
	}

#ifdef ULTRADEBUG
	PRINTK("%s: found caller id signature: ",__FUNCTION__);
	{
		int i;
		for(i=0;i<CALLER_ID_LEN && clrid[i]!=0;i++)
		{
			PRINTK("%c",clrid[i]);
		}
		PRINTK("\n");
	}
#endif
	// copy the caller id and exit.
	strncpy(id, clrid, CALLER_ID_LEN);

	free_page((unsigned long)tmppage);
	return 0;

fail0:
	free_page((unsigned long)tmppage);
	return -1;
}

/*******************************************************************************/

// TODO: support offset
static int proc_read_challenge(char *buf, char **start, off_t offset, int size,
                               int *eof, void *data)
{
	int res;
	char callerid[CALLER_ID_LEN];
	int rval;
	unsigned long flags;
	struct auth_list_elem *authelem;

	if(size<CHALLENGE_LEN)
	{
		printk(KERN_ERR "%s: can't output challenge: outbuffer too small\n",
				__FUNCTION__);
		return 0;
	}

	/*
	 * find callerid in the process' environment
	 */
	if((res=get_callerid_from_environment(current, callerid))<0)
	{
		PRINTK("%s: error getting caller id\n",__FUNCTION__);
		return 0;
	}

	rval=0;

	/*
	 * search the id in the infolist
	 */
	spin_lock_irqsave(&auth_list_lock, flags);
	list_for_each_entry(authelem, &auth_list, list)
	{
		if(memcmp(authelem->callerid, callerid, CALLER_ID_LEN)==0)
		{
			/*
			 * got the right entry. copy the challenge.
			 */
			memcpy(buf, authelem->challenge, CHALLENGE_LEN);
			rval=CHALLENGE_LEN;
			break;
		}
	}
	spin_unlock_irqrestore(&auth_list_lock, flags);

	if(rval==0)
	{
		printk(KERN_ERR "%s: couldn't find list element\n",__FUNCTION__);
	}
	return rval;
}

/*******************************************************************************/

// TODO: support offset
static int proc_read_signature(char *buf, char **start, off_t offset, int size,
                               int *eof, void *data)
{
	int res;
	char callerid[CALLER_ID_LEN];
	int rval;
	unsigned long flags;
	struct auth_list_elem *authelem;

	/*
	 * find callerid in the process' environment
	 */
	if((res=get_callerid_from_environment(current, callerid))<0)
	{
		PRINTK("%s: error getting caller id\n",__FUNCTION__);
		return 0;
	}

	rval=0;

	/*
	 * search the id in the infolist
	 */
	spin_lock_irqsave(&auth_list_lock, flags);
	list_for_each_entry(authelem, &auth_list, list)
	{
		if(memcmp(authelem->callerid, callerid, CALLER_ID_LEN)==0)
		{
			/*
			 * got the right entry. copy the signature
			 */
			if(size<authelem->siglen)
			{
				printk(KERN_ERR "%s: can't output signature: outbuffer too small\n",
						__FUNCTION__);
				spin_unlock_irqrestore(&auth_list_lock, flags);
				return 0;
			}

			memcpy(buf, authelem->signature, authelem->siglen);
			rval=authelem->siglen;
			break;
		}
	}
	spin_unlock_irqrestore(&auth_list_lock, flags);

	if(rval==0)
	{
		printk(KERN_ERR "%s: couldn't find list element\n",__FUNCTION__);
	}
	return rval;
}
/*******************************************************************************/

#if LINUX_VERSION_CODE < LinuxVersionCode(2,5,0)
#define __user
#endif

static int proc_write_result(struct file *file, const char __user *buf, unsigned long count,
                             void *data)
{
	int res;
	char callerid[CALLER_ID_LEN];
	int rval;
	unsigned long flags;
	struct auth_list_elem *authelem;
	size_t len;
	unsigned char *ptr;

	/*
	 * find callerid in the process' environment
	 */
	if((res=get_callerid_from_environment(current, callerid))<0)
	{
		PRINTK("%s: error getting caller id\n",__FUNCTION__);
		return -EPERM;
	}

	/*
	 * read in userspace data
	 */
	ptr=(unsigned char *)kmalloc(count, GFP_KERNEL);
	if(ptr==NULL)
	{
		printk(KERN_ERR "%s: out of memory\n",__FUNCTION__);
		return -ENOMEM;
	}
	if(copy_from_user(ptr, buf, count))
	{
		printk(KERN_ERR "%s: couldn't copy result from userland\n",__FUNCTION__);
		kfree(ptr);
		return -EIO;
	}

	rval=0;

	/*
	 * search the id in the infolist
	 */
	spin_lock_irqsave(&auth_list_lock, flags);
	list_for_each_entry(authelem, &auth_list, list)
	{
		if(memcmp(authelem->callerid, callerid, CALLER_ID_LEN)==0)
		{
			/*
			 * got the right entry. store the result.
			 */
			size_t rest;
			if(sizeof(authelem->result)<=authelem->resultlen)
			{
				printk(KERN_ERR "%s: internal error: resultlen > sizeof(result) (%d > %d)\n",
						__FUNCTION__, authelem->resultlen, sizeof(authelem->result));
				spin_unlock_irqrestore(&auth_list_lock, flags);
				kfree(ptr);
				return -EIO;
			}
			/*
			 * space left in the buffer
			 */
			rest=sizeof(authelem->result)-authelem->resultlen;
			rest--;   // leave space for \0

			len=(count < rest ? count : rest);

			/*
			 * copy data
			 */
			memcpy(authelem->result+authelem->resultlen, ptr, len);
			authelem->resultlen+=len;
			authelem->result[authelem->resultlen]=0;

			rval=len;
			break;
		}
	}
	spin_unlock_irqrestore(&auth_list_lock, flags);

	kfree(ptr);

	if(rval==0)
	{
		printk(KERN_ERR "%s: couldn't find list element\n",__FUNCTION__);
		rval=-1;
	}
	return rval;
}
/*******************************************************************************/
/* 
 * Calculates an UDP checksum.
 * u16 buff[] is an array containing all the octets in the UDP header and data.
 * u16 len is the length (number of octets) of the UDP header and data.
 * u16 src_addr and u16 dest_addr are the IP source and destination addresses (network byte order)
 * returns the checksum (in host byte order)
 */
static __u16 udp_sum_calc(__u16 len, __u32 saddr,__u32 daddr, __u8 buff[])
{
	__u16 word16;
	__u32 sum; 
	__u16 src_addr[4], dest_addr[4];
	int i;

	sum=0;

	src_addr[3]=(saddr>>24)&0xFF;
	src_addr[2]=(saddr>>16)&0xFF;
	src_addr[1]=(saddr>>8)&0xFF;
	src_addr[0]=saddr&0xFF;
	dest_addr[3]=(daddr>>24)&0xFF;
	dest_addr[2]=(daddr>>16)&0xFF;
	dest_addr[1]=(daddr>>8)&0xFF;
	dest_addr[0]=daddr&0xFF;

	/*
	 * make 16 bit words out of every two adjacent 8 bit words and 
	 * calculate the sum of all 16 bit words
	 */
	for (i=0;i<len-1;i=i+2){
		__u16 b1, b2;
		b1=(__u16)buff[i];
		b2=(__u16)buff[i+1];
		word16 =((b1<<8)&0xFF00)+(b2&0xFF);
		sum = sum + (__u32)word16;
	}
	/*
	 * check if padding is needed
	 */
	if(len%2) 
	{
		word16 =((buff[i]<<8)&0xFF00)+(0&0xFF);
		sum = sum + (__u32)word16;
	}
	/*
	 * add the TCP pseudo header which contains:
	 * the IP source and destination addresses,
	 */
	for (i=0;i<4;i=i+2){
		word16 =((src_addr[i]<<8)&0xFF00)+(src_addr[i+1]&0xFF);
		sum=sum+word16;   
	}
	for (i=0;i<4;i=i+2){
		word16 =((dest_addr[i]<<8)&0xFF00)+(dest_addr[i+1]&0xFF);
		sum=sum+word16;   
	}
	/*
	 * the protocol number and the length of the TCP packet
	 */
	sum = sum + IPPROTO_UDP + len;

	/*
	 * keep only the last 16 bits of the 32 bit calculated sum and add the carries
	 */
	while (sum>>16)
		sum = (sum & 0xFFFF)+(sum >> 16);

	/* 
	 * Take the one's complement of sum
	 */
	sum = ~sum;

	return ((__u16) sum);
}

/*******************************************************************************/

/*
 * sends out a reply packet
 * doesn't sleep
 */
void send_udp_packet(__u32 saddr, __u32 daddr, __u16 source, __u16 dest, 
           char proto, char *syncookie, char *payload, size_t payloadlen)
{
	struct sk_buff *newskb;
	struct iphdr *iph;
	struct udphdr *udph;
	unsigned char *ptr;

	/*
	 * Allocate sk_buff
	 */
	newskb=dev_alloc_skb(sizeof(struct iphdr)+sizeof(struct udphdr)+
			1+SYNCOOKIE_HASH_SIZE+payloadlen);
	if(newskb==NULL)
	{
		printk(KERN_ERR "%s: dev_alloc_skb returned NULL",__FUNCTION__);
		return;
	}

	ptr=skb_put(newskb, sizeof(struct iphdr));
	newskb->nh.iph=iph=(struct iphdr *)ptr;
	ptr=skb_put(newskb, sizeof(struct udphdr));
	newskb->h.uh=udph=(struct udphdr *)ptr;

	iph->version=4;
	iph->ihl=sizeof(struct iphdr)/4;
	iph->tos=0;
	iph->id=0;
	iph->frag_off=0;
	iph->ttl=255;
	iph->protocol=IPPROTO_UDP;
	iph->tot_len=htons(sizeof(struct iphdr)+sizeof(struct udphdr)+1+
			SYNCOOKIE_HASH_SIZE+payloadlen);
	iph->saddr=saddr;
	iph->daddr=daddr;

	udph->source=source;
	udph->dest=dest;
	udph->len=htons(sizeof(struct udphdr)+1+SYNCOOKIE_HASH_SIZE+payloadlen);

	/*
	 * copy the syncookie and the payload
	 */
	ptr=skb_put(newskb,1);
	ptr[0]=proto;
	ptr=skb_put(newskb,SYNCOOKIE_HASH_SIZE);
	memcpy(ptr,syncookie,SYNCOOKIE_HASH_SIZE);
	if(payload!=NULL && payloadlen>0)
	{
		ptr=skb_put(newskb,payloadlen);
		memcpy(ptr,payload,payloadlen);
	}

	/*
	 * get checksums right
	 */
	iph->check=0;
	iph->check=ip_fast_csum((void *)iph,iph->ihl);
	udph->check=0;
	udph->check=htons(udp_sum_calc(sizeof(struct udphdr)+1+SYNCOOKIE_HASH_SIZE+
				payloadlen, iph->saddr, iph->daddr, (__u8 *)udph));
	{
		/*
		 * reroute packet and send it out
		 */
		struct rtable *rt;
#if LINUX_VERSION_CODE < LinuxVersionCode(2,5,0)
		struct rt_key key = {};
		key.dst=iph->daddr;
		key.src=iph->saddr;
		key.tos=RT_TOS(iph->tos)|RTO_CONN;
		if(ip_route_output_key(&rt,&key)!=0)
		{
			printk(KERN_ERR "%s: ip_route_output_key returned non-zero value\n",__FUNCTION__);
			return;
		}
#else
		struct flowi fl={.nl_u={.ip4_u=
			{
				.daddr=iph->daddr,
				.tos=(RT_TOS(iph->tos)|RTO_CONN)
			}}};
		if(ip_route_output_key(&rt, &fl)!=0)
		{
			printk(KERN_ERR "%s: ip_route_output_key returned non-zero value\n",__FUNCTION__);
			return;
		}
#endif
		newskb->dst=&(rt->u.dst);
		my_dst_output(newskb);
	}
}

/*******************************************************************************/

#if LINUX_VERSION_CODE < LinuxVersionCode(2,5,0)

struct gpg_call_info
{
	char *arg0;
	char **argv;
	char **envp;
};

static int __my_call_usermodehelper_wait(void *arg)
{
	struct gpg_call_info *info=(struct gpg_call_info *)arg;

	// do it
	exec_usermodehelper(info->arg0, info->argv, info->envp);
	return 0;
}

/*
 * my own version of call_usermodehelper for v2.4-kernels, waits for
 * completion of the invoked process.
 * returns <0 on error, >=0 on success
 */
static int my_call_usermodehelper_wait(char *arg0, char **argv, char **envp)
{
	int pid;
	struct gpg_call_info info;

	info.arg0=arg0;
	info.argv=argv;
	info.envp=envp;

	pid=kernel_thread(__my_call_usermodehelper_wait, (void *)&info, SIGCHLD);
	if(pid>0)
	{
		int i, res;
		mm_segment_t oldfs;

		while(1)
		{
			oldfs=get_fs();
			set_fs(get_ds());
			res=waitpid(-1, &i, 0);
			set_fs(oldfs);
			if(pid==res)
				break;
			PRINTK(KERN_ERR "%s: waitpid returned %d (errno=%d)\n",__FUNCTION__,res,errno);
			yield();
		}
	}
	return pid;
}
#endif

/*******************************************************************************/

/*
 * Make a copy of the auth_list.
 * Assumes that the auth_list_lock is already held (and thus doesn't sleep).
 * \param copy Pointer to the list head of the new copy.
 * \param clean_auth_result Clean the result field of the auth element
 * \returns 0 on success, <0 on error. Leaves everything allocated and doesn't
 * attempt to clean anything up.
 */
static int copy_auth_list(struct list_head *copy, int clean_auth_result)
{
	struct auth_list_elem *authelem, *auth;
	INIT_LIST_HEAD(copy);

	list_for_each_entry(authelem, &auth_list, list)
	{
		auth=(struct auth_list_elem *)kmalloc(sizeof(struct auth_list_elem),
				GFP_ATOMIC);
		if(auth==NULL)
		{
			printk("sigknock: %s: out of memory!\n",__FUNCTION__);
			return -ENOMEM;
		}

		if(clean_auth_result)
		{
			memset(authelem->result, 0, sizeof(authelem->result));
			authelem->resultlen=0;
		}

		*auth=*authelem;

		list_add(&auth->list, copy);
	}

	return 0;
}

/*******************************************************************************/

static int signature_thread(void *dummy)
{
	char *cmdline, *buffer;
	DECLARE_WAITQUEUE(wait, current);

#if LINUX_VERSION_CODE < LinuxVersionCode(2,5,0)
	daemonize();
#else
	daemonize("ksigknock");
#endif

	add_wait_queue(&signature_wait_queue, &wait);
	exit_signature_thread_command=0;

	cmdline=(char *)__get_free_page(GFP_KERNEL);
	buffer=(char *)__get_free_page(GFP_KERNEL);
	if(cmdline==NULL || buffer==NULL)
	{
		printk(KERN_ERR "%s: out of memory!\n",__FUNCTION__);
		if(cmdline) free_page((unsigned long)cmdline);
		if(buffer) free_page((unsigned long)buffer);
		return -ENOMEM;
	}

	PRINTK(KERN_INFO "signature thread ready\n");

	do
	{
		int res;
		int size;
		char *argv[]={"/bin/bash","-c",NULL,NULL};
		char *envp[]={NULL, NULL};
		char *ptr;
		unsigned long flags;
		unsigned char callerid[CALLER_ID_LEN+1];   // +1 for trailing \0
		struct auth_list_elem *authelem, *localauthelem;
		struct info_list_elem *infoelem;
		LIST_HEAD(local_auth_list);

		/*
		 * wait for the netfilter hook to wake us up
		 */
		sleep_on(&signature_wait_queue);
		PRINTK("signature thread woke up\n");

		if(exit_signature_thread_command==1)
			continue;

		res=0;

		/*
		 * copy the whole list locally, so we don't need to hold the
		 * spinlocks the whole time (would be impossible anyway because
		 * we go to sleep). 
		 */
		spin_lock_irqsave(&auth_list_lock, flags);
		copy_auth_list(&local_auth_list, 1);
		spin_unlock_irqrestore(&auth_list_lock, flags);

		/*
		 * now search for the auth elements that need_check
		 */
		list_for_each_entry(localauthelem, &local_auth_list, list)
		{
			if(localauthelem->needs_check==1)
			{
				/* 
				 * got it!
				 */
				localauthelem->needs_check=0;

				/*
				 * invoke gpg
				 */
				snprintf(cmdline,PAGE_SIZE,"%s --homedir %s --verify %s %s > %s 2>&1",
						GPG_PATH, SIGKNOCK_HOMEDIR,
						"/proc/net/ipt_sigknock/signature",
						"/proc/net/ipt_sigknock/challenge",
						"/proc/net/ipt_sigknock/result");
				memset(callerid, 0, sizeof(callerid));
				memcpy(callerid, localauthelem->callerid, CALLER_ID_LEN);
				snprintf(buffer,PAGE_SIZE,"%s%s",CALLER_ID_SIGNATURE, callerid);

				argv[2]=cmdline;
				envp[0]=buffer;

#if LINUX_VERSION_CODE < LinuxVersionCode(2,5,0)
				res=my_call_usermodehelper_wait(*argv, argv, envp);
#else
				res=call_usermodehelper(*argv, argv, envp, 1);
#endif
				if(res<0)
				{
					printk(KERN_ERR "%s: error running gpg: %d\n",__FUNCTION__,res);
					continue;
				}

				/*********
				 * 
				 * parse result 
				 * 
				 *********/

				/*
				 * search for the element again and read the result.
				 */
				res=0;
				spin_lock_irqsave(&auth_list_lock, flags);
				list_for_each_entry(authelem, &auth_list, list)
				{
					if(localauthelem->saddr==authelem->saddr)
					{
						res=1;
						memset(buffer, 0, PAGE_SIZE);
						size=(authelem->resultlen > (PAGE_SIZE-1) ? (PAGE_SIZE-1) : authelem->resultlen);
						memcpy(buffer, authelem->result, size);
						/* 
						 * buffer is guaranteed to be \0-terminated.
						 */
						break;
					}
				}
				spin_unlock_irqrestore(&auth_list_lock, flags);

				if(res==0)
				{
					printk(KERN_ERR "%s: unable to find my list element!\n",__FUNCTION__);
					continue;
				}

				/*
				 * get the UID from the gpg-output-string,
				 * then run through all info-elements and set every entry to authorized
				 * where the uids match
				 */
				ptr=strstr(buffer, "gpg: Good signature from \"");
				if(ptr)
				{
					int i;
					/*
					 * extract the uid
					 */
					ptr+=strlen("gpg: Good signature from \"");
					memset(cmdline, 0, PAGE_SIZE);
					for(i=0;i<PAGE_SIZE-1 && ptr[i]!=0 && ptr[i]!='"'; i++)
						cmdline[i]=ptr[i];

					printk(KERN_INFO "sigknock: GOOD signature for uid '%s' "
							"from %u.%u.%u.%u\n", cmdline, ADDRARGS(authelem->saddr));

					spin_lock_irqsave(&info_list_lock, flags);
					list_for_each_entry(infoelem, &info_list, list)
					{
						int i;
						for(i=0;i<infoelem->pkinfo.uid_count && 
								i<sizeof(infoelem->pkinfo.uid)/sizeof(infoelem->pkinfo.uid[0]);++i)
						{
							if(strcmp(infoelem->pkinfo.uid[i], cmdline)==0)
							{
								/*
								 * match!
								 */
								infoelem->authorized=1;
								infoelem->timestamp=jiffies;
							}
						}
					}
					spin_unlock_irqrestore(&info_list_lock, flags);

					/*
					 * update the real auth element
					 */
					spin_lock_irqsave(&auth_list_lock, flags);
					list_for_each_entry(authelem, &auth_list, list)
					{
						if(authelem->saddr==localauthelem->saddr)
						{
							authelem->needs_check=0;

							/*
							 * invalidate the challenge to prevent a simple replay attack
							 */
							memset(authelem->challenge, 0, sizeof(authelem->challenge));

							/*
							 * clean up the result, just to be sure
							 */
							memset(authelem->result, 0, sizeof(authelem->result));
							authelem->resultlen=0;
							break;
						}
					}
					spin_unlock_irqrestore(&auth_list_lock, flags);
				}
				else
				{
					/*
					 * nothing found, deny
					 */
					printk(KERN_INFO "sigknock: BAD signature request "
							"from %u.%u.%u.%u\n", ADDRARGS(authelem->saddr));
					PRINTK(KERN_INFO "sigknock: gpg output was:\n%s\n",buffer);
				}
			} // if(localauthelem->needs_check)
		} // list_for_each_entry(localauthelem)

		/*
		 * free the local copy
		 */
cleanup1: 
		list_for_each_entry(localauthelem, &local_auth_list, list)
		{
			list_del(&localauthelem->list);
			kfree(localauthelem);
			goto cleanup1;
		}
#ifdef DEBUG
		spin_lock_irqsave(&info_list_lock, flags);
		dump_infolist(&info_list);
		spin_unlock_irqrestore(&info_list_lock, flags);
		spin_lock_irqsave(&auth_list_lock, flags);
		dump_authlist(&auth_list);
		spin_unlock_irqrestore(&auth_list_lock, flags);
#endif
	} while(exit_signature_thread_command==0);

	free_page((unsigned long)cmdline);
	free_page((unsigned long)buffer);
	cmdline=buffer=NULL;

	/*
	 * up the exit semaphore so cleanup_module knows that we are finished
	 */
	up(&sem_exit_notify);

	PRINTK("%s: terminating\n",__FUNCTION__);

	yield();

	return 0;
}

/*******************************************************************************/

/*
 * needs lock on info_list_lock
 */
static void dump_infolist(struct list_head *l)
{
	struct info_list_elem *elem;

	PRINTK("****************************\n");
	PRINTK("dumping infolist:\n");
	list_for_each_entry(elem, l, list)
	{
		int i;
		PRINTK("   PKINFO:\n");
		PRINTK("      timeout: %d%c\n",elem->pkinfo.timeout, elem->pkinfo.timeout_char);
		PRINTK("      uid: ");
		for(i=0;i<elem->pkinfo.uid_count;++i)
			PRINTK("%s, ",elem->pkinfo.uid[i]);
		PRINTK("\n");
		PRINTK("      close-after-syn: %d\n", elem->pkinfo.close_after_syn);
		PRINTK("   authorized: %d\n",elem->authorized);
		PRINTK("   timestamp: %ld\n",elem->timestamp);
		PRINTK("   timediff: %ld\n",elem->timediff);
	}
	PRINTK("****************************\n");
}

/*******************************************************************************/

/*
 * needs lock on auth_list_lock
 */
static void dump_authlist(struct list_head *l)
{
	struct auth_list_elem *auth;
	PRINTK("****************************\n");
	PRINTK("dumping authlist:\n");
	list_for_each_entry(auth, &auth_list, list)
	{
		int i;
		PRINTK("     saddr: %u.%u.%u.%u\n",ADDRARGS(auth->saddr));
		PRINTK("     needs_check: %d\n",auth->needs_check);
		PRINTK("     challenge: ");
		for(i=0;i<CHALLENGE_LEN;i++)
			PRINTK("%02x ",auth->challenge[i]);
		PRINTK("\n     challenge_timestamp: %ld\n",auth->challenge_timestamp);      
		PRINTK("     signature: ");
		for(i=0;i<auth->siglen;i++)
			PRINTK("%02x ",auth->signature[i]);
		PRINTK("\n     siglen: %d\n",auth->siglen);      
		PRINTK("     callerid: ");
		for(i=0;i<CALLER_ID_LEN;i++)
			PRINTK("%02x ",auth->callerid[i]);
		PRINTK("\n     result: ");
		for(i=0;i<auth->resultlen;i++)
			PRINTK("%02x ",auth->result[i]);
		PRINTK("\n     resultlen: %d\n",auth->resultlen);
		PRINTK("     - - - - - - - - - - - - - - - - - \n");
	}
	PRINTK("****************************\n");
}

/*******************************************************************************/

/*
 * This hook is called when there is a packet ready on the SIGKNOCK-KEYPORT
 */
static unsigned int ipt_pk_target(struct sk_buff **skb,
#if LINUX_VERSION_CODE < LinuxVersionCode(2,5,0)
	       unsigned int hooknum,
#endif
	       const struct net_device *in,
	       const struct net_device *out,
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,5,0)
	       unsigned int hooknum,
#endif
	       const void *targinfo,
	       void *userinfo)
{
	struct iphdr *iph = NULL;
	struct udphdr *udph=NULL;
	struct sk_buff *pskb;
	struct syncookie_struct syncookie, old_syncookie;
	unsigned char syncookie_hash[SYNCOOKIE_HASH_SIZE];
	unsigned char old_syncookie_hash[SYNCOOKIE_HASH_SIZE];

	pskb=*skb;

	if(pskb->protocol!=ntohs(ETH_P_IP))
	{
		PRINTK("sigknock: MAC layer reports this is not an IP packet - ignoring\n");
		return NF_ACCEPT;
	}

	iph=pskb->nh.iph;

	if(iph->protocol==IPPROTO_UDP)
	{
		size_t payloadsize;
		unsigned char *payloadptr;

		udph=(struct udphdr *)(pskb->data+iph->ihl*4);

		/*
		 * check if the sizes in the packet are sane
		 */
		if(pskb->len!=ntohs(iph->tot_len) || pskb->len!=iph->ihl*4+ntohs(udph->len))
		{
			printk("sigknock: malformed packet from %u.%u.%u.%u, ignoring\n", ADDRARGS(iph->saddr));
			return NF_ACCEPT;
		}

		payloadsize=pskb->len-iph->ihl*4-sizeof(struct udphdr);
		payloadptr=((unsigned char *)udph)+sizeof(struct udphdr);

		PRINTK("%s: hit on the keyport\n",__FUNCTION__);

		/*
		 * Generate the syncookie for this packet - never a bad idea :)
		 */
		memset(&syncookie, 0, sizeof(syncookie));
		syncookie.timestamp=jiffies>>jiffies_shift_for_delay;
		syncookie.sigthread_pid=sigthread_pid;
		syncookie.saddr=iph->saddr;
		syncookie.daddr=iph->daddr;
		syncookie.sport=udph->source;
		syncookie.dport=udph->dest;
		SHA1HashLen((unsigned char *)&syncookie, sizeof(syncookie), syncookie_hash);

		/*
		 * Also generate the syncookie from the previous time period. 
		 */
		memset(&old_syncookie, 0, sizeof(old_syncookie));
		old_syncookie=syncookie;
		old_syncookie.timestamp--;
		SHA1HashLen((unsigned char *)&old_syncookie, sizeof(old_syncookie), old_syncookie_hash);

		if(payloadsize==0)
		{
			PRINTK("%s: not enough payload\n",__FUNCTION__);
			return NF_ACCEPT;
		}
		else if(payloadsize==1 && payloadptr[0]==COOKIE_REQUEST)
		{
			/*
			 * The client wants us to send a SYN-cookie
			 */
			send_udp_packet(iph->daddr, iph->saddr, udph->dest, udph->source,
					COOKIE_RESPONSE, syncookie_hash, NULL, 0);

			return NF_DROP;
		}
		else if(payloadsize>=SYNCOOKIE_HASH_SIZE+1) // [1]
		{
			/*
			 * check the syncookie!
			 */
			if(memcmp(syncookie_hash, payloadptr+1, SYNCOOKIE_HASH_SIZE))
			{
				/*
				 * OK that didn't match. Also check the old syncookie from the previous
				 * time period, it could be that the syncookie time just wrapped between the
				 * two requests from the client.
				 */
				if(memcmp(old_syncookie_hash, payloadptr+1, SYNCOOKIE_HASH_SIZE))
				{
					printk("sigknock: syncookies don't match in packet from %u.%u.%u.%u\n",
							ADDRARGS(iph->saddr));
					return NF_ACCEPT;
				}
#ifdef DEBUG
				else
				{
					printk("sigknock: first syncookie match failed, old syncookie matched. "
							"should be OK. saddr=%u.%u.%u.%u\n",ADDRARGS(iph->saddr));
				}
#endif
			}

			/*
			 * at this point we can be sure that the packet isn't spoofed,
			 * thanks to the syncookie.
			 */

			if(payloadptr[0]==CHALLENGE_REQUEST && payloadsize==1+SYNCOOKIE_HASH_SIZE)
			{
				/*
				 * the client requests a challenge
				 */
				unsigned char chg[CHALLENGE_LEN];
				struct auth_list_elem *authelem, *auth, *oldcandidate;
				int i=0;

				// TODO: does that function really not sleep?
				get_random_bytes(chg,sizeof(chg));
#ifdef DEBUG
				PRINTK("%s: Generated challenge = ",__FUNCTION__);
				for(i=0;i<sizeof(chg)/sizeof(chg[0]);++i)
					PRINTK("%02x ",chg[i]);
				PRINTK("\n");
#endif

				/*
				 * store the challenge and the source address in the auth list.
				 * first, search if there is already such an element. If not,
				 * reuse a slot of an old entry.
				 */
				auth=oldcandidate=NULL;
				spin_lock(&auth_list_lock);
				list_for_each_entry(authelem, &auth_list, list)
				{
					if(authelem->saddr==iph->saddr)
					{
						/*
						 * got the entry with the same source address.
						 */
						auth=authelem;
						PRINTK("%s: found an auth entry with the same source "
								"address, using it.\n",__FUNCTION__);
						break;
					}
					else if(jiffies-authelem->challenge_timestamp > 30*HZ)
					{
						/*
						 * found an old entry. Old entries are most likely
						 * alread processed, but still there because we keep them
						 * in the list so we can throttle the amount of requests
						 * per second made by clients. This is there to prevent
						 * DoS.
						 */
						oldcandidate=authelem;
						PRINTK("%s: found an old entry for the auth element, using it.\n",
								__FUNCTION__);
					}
				}
				if(auth==NULL)
				{
					if(oldcandidate==NULL)
					{
						/*
						 * allocate a new one
						 */
						auth=(struct auth_list_elem *)kmalloc(sizeof(struct auth_list_elem),
								GFP_ATOMIC);
						PRINTK("kmalloc %p\n",auth);
						if(auth==NULL)
						{
							printk(KERN_ERR "sigknock: can't allocate new memory, dropping packet\n");
							spin_unlock(&auth_list_lock);
							return NF_ACCEPT;
						}
						list_add(&auth->list, &auth_list);
					}
					else
					{
						/* 
						 * use the old slot
						 */
						auth=oldcandidate;
					}
				}
				else
				{
					/*
					 * knock-throttling 
					 */
					if(jiffies-auth->challenge_timestamp < KNOCK_THRESHOLD*HZ)
					{
						printk("sigknock: host %u.%u.%u.%u is knocking too fast (%ds < %ds), "
								"ignoring\n",ADDRARGS(iph->saddr), 
								(int)(auth->challenge_timestamp-jiffies)/HZ, KNOCK_THRESHOLD);
						PRINTK("jiffies = %ld, challenge_timestamp = %ld\n",
								jiffies,auth->challenge_timestamp);
						spin_unlock(&auth_list_lock);
						return NF_DROP;
					}
				}
				memcpy(auth->challenge,chg,CHALLENGE_LEN);
				auth->saddr=iph->saddr;
				auth->challenge_timestamp=jiffies;
				auth->signature_timestamp=0;
				memset(auth->signature, 0, sizeof(auth->signature));
				auth->siglen=0;
				memset(auth->callerid, 0, sizeof(auth->callerid));
				memset(auth->result, 0, sizeof(auth->result));
				auth->resultlen=0;
				auth->needs_check=0;

				/*
				 * done with the auth info
				 */
				spin_unlock(&auth_list_lock);

				/*
				 * send back reply
				 */
				send_udp_packet(iph->daddr, iph->saddr, udph->dest, udph->source,
						CHALLENGE_RESPONSE, syncookie_hash, chg, sizeof(chg));

				return NF_DROP;
			}
			else if(payloadptr[0]==SIGNATURE_REQUEST)
			{
				/*
				 * the client returns a signed challenge
				 */
				unsigned char *sigptr;
				size_t sigsize;
				int i;
				struct auth_list_elem *authelem, *auth;

				sigsize=payloadsize-1-SYNCOOKIE_HASH_SIZE;   // integer overflow checked above [1]
				if(sigsize>sizeof(authelem->signature))
				{
					printk(KERN_NOTICE "%s: malformed packet, signature payload too large, ignoring\n",__FUNCTION__);
					return NF_ACCEPT;
				}
				sigptr=payloadptr+1+SYNCOOKIE_HASH_SIZE;

				/*
				 * get the authinfo struct that should be there for this source
				 * address.
				 */
				auth=NULL;
				spin_lock(&auth_list_lock);
				list_for_each_entry(authelem, &auth_list, list)
				{
					if(authelem->saddr==iph->saddr)
					{
						/*
						 * got the entry with the same source address.
						 */
						auth=authelem;
						break;
					}
				}
				if(auth==NULL)
				{
					/*
					 * whoops! we know nothing about this host, at least we didn't issue
					 * him a challenge. drop
					 */
					printk("sigknock: host %u.%u.%u.%u is sending a signature, but I didn't "
							"send him a challenge! ignoring\n",ADDRARGS(iph->saddr));
					spin_unlock(&auth_list_lock);
					return NF_ACCEPT;
				}
				/* 
				 * knock-throttling 
				 */
				if(jiffies-auth->signature_timestamp < KNOCK_THRESHOLD*HZ)
				{
					printk("sigknock: host %u.%u.%u.%u is knocking too fast (%ds < %ds), "
							"ignoring (signature request)\n",ADDRARGS(iph->saddr), 
							(int)(auth->signature_timestamp-jiffies)/HZ, KNOCK_THRESHOLD);
					spin_unlock(&auth_list_lock);
					return NF_DROP;
				}
				auth->signature_timestamp=jiffies;

				/*
				 * copy the info from the packet
				 */
				memcpy(auth->signature, sigptr, sigsize);
				auth->siglen=sigsize;

				/*
				 * generate a caller id and store it
				 */
				get_random_bytes(auth->callerid, CALLER_ID_LEN);
				for(i=0; i<CALLER_ID_LEN; i++)
				{
					auth->callerid[i]=(auth->callerid[i]%26)+'A';
				}
				auth->callerid[CALLER_ID_LEN]=0;

#ifdef DEBUG
				printk("%s: got signature request:\n",__FUNCTION__);
				for(i=0;i<sigsize;i++)
				{
					printk("%02x ",sigptr[i]);
				}
				printk("\n");
#endif
				auth->needs_check=1;

				/*
				 * we are done with info from the list
				 */
				spin_unlock(&auth_list_lock);

				/*
				 * Wake up the signature checking thread so it can do its work.
				 */
				wake_up(&signature_wait_queue);

				return NF_ACCEPT;
			}
			else
			{
				PRINTK("%s: protocol error\n",__FUNCTION__);
			}
		}
		else
		{
			PRINTK("%s: payloadsize too small...\n",__FUNCTION__);
		}
	}

	return NF_ACCEPT;
}

/*******************************************************************************/

/*
 * The match function - returns non-zero if the packet matches
 * called as softirq
 */
static int ipt_pk_match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
#if LINUX_VERSION_CODE < LinuxVersionCode(2,5,0)
      const void *hdr,
      u_int16_t datalen,
#endif
      int *hotdrop)
{
	struct info_list_elem *elem, *info;
	const struct ipt_sigknock_info *i = matchinfo;
	int rval;

	info=NULL;

	/*
	 * we hold that spinlock during the whole procedure. 
	 * That's fine, we never sleep.
	 */
	spin_lock(&info_list_lock);
	list_for_each_entry(elem, &info_list, list)
	{
		if(infocmp(&elem->pkinfo, (struct ipt_sigknock_info *)i)==0)
		{
			info=elem;
			break;
		}
	}

	if(info==NULL)
	{
		printk(KERN_ERR "%s: can't find my entry!\n",__FUNCTION__);
		rval=0;
	}
	else
	{
		int i;
		if(info->authorized && info->pkinfo.timeout!=-1 && info->timestamp+info->timediff<jiffies)
		{
			/*
			 * Timeout ran out! Close the port.
			 */
			PRINTK("sigknock: timeout ran out, closing port\n");
			info->authorized=0;
		}

		PRINTK("sigknock: %s: %s match! uid = ",__FUNCTION__,info->authorized==0 ? "denying" : "allowing");
		for(i=0;i<info->pkinfo.uid_count;++i)
			PRINTK("%s, ",info->pkinfo.uid[i]);

		/*
		 * if we can, print out the source address
		 */
		{
			if(skb->protocol==ntohs(ETH_P_IP))
			{
				struct iphdr *iph=skb->nh.iph;
				PRINTK("source addr = %u.%u.%u.%u", ADDRARGS(iph->saddr));
			}
		}

		PRINTK("\n");

		rval=info->authorized;

		if(info->authorized && info->pkinfo.close_after_syn)
		{
			/*
			 * check if this is a SYN packet, if so, close the match ...
			 */
			struct iphdr *iph=NULL;
			struct tcphdr *tcph=NULL;

			iph=skb->nh.iph;
			if(iph->protocol==IPPROTO_TCP)
			{
				tcph=(struct tcphdr *)(skb->data+(skb->nh.iph->ihl*4));
				if(tcph->syn && !tcph->ack && !tcph->rst)        // SYN packets
				{
					/*
					 * that's really a SYN packet. Deny further matches!
					 */
					info->authorized=0;
					PRINTK("%s: got a SYN packet and close_after_syn is true! Doing so!\n",
							__FUNCTION__);
				}
			}
		}
	}

	spin_unlock(&info_list_lock);

	return rval;
}

/*******************************************************************************/

// Checks an iptables-entry for correctness - returns 0 if it is invalid
static int ipt_pk_checkentry_target(const char *tablename,
		   const struct ipt_entry *e,
		   void *targinfo,
		   unsigned int matchsize,
		   unsigned int hook_mask)
{
	return 1;
}

/*******************************************************************************/

// Checks an iptables-entry for correctness - returns 0 if it is invalid
static int ipt_pk_checkentry_match(const char *tablename,
		   const struct ipt_ip *ip,
		   void *matchinfo,
		   unsigned int matchsize,
		   unsigned int hook_mask)
{
	const struct ipt_sigknock_info *info = matchinfo;
	struct info_list_elem *elem;
	unsigned long flags;
	static int count=0;

	count++;

	/* we can use a sigknock rule just about anywhere
	 * if (hook_mask
	 & ~((1 << NF_IP_LOCAL_IN) )) {
	 printk("ipt_sigknock: only valid for LOCAL_IN.\n");
	 return 0;
	 } */

	if (matchsize != IPT_ALIGN(sizeof(struct ipt_sigknock_info)))
		return 0;

#ifdef ULTRADEBUG
	// lacks UID adaption
	/*   PRINTK("%s: adding iptables entry:\n",__FUNCTION__);
	     PRINTK("   timeout: %d%c\n   uid: %s\n   close_after_syn: %d\n",
	     info->timeout, info->timeout_char, info->uid, info->close_after_syn);*/
#endif

	elem=(struct info_list_elem *)kmalloc(sizeof(struct info_list_elem), GFP_KERNEL);
	if(elem==NULL)
	{
		printk(KERN_ERR "sigknock: %s: out of memory\n",__FUNCTION__);
		return 0;
	}
	PRINTK("** %s (%d): kmalloc: %p\n",__FUNCTION__, count, elem);

	elem->pkinfo=*info;
	if(elem->pkinfo.timeout!=-1)
	{
		switch(elem->pkinfo.timeout_char)
		{
			case 's': elem->timediff=elem->pkinfo.timeout*HZ; break;
			case 'm': elem->timediff=elem->pkinfo.timeout*60*HZ; break;
			case 'h': elem->timediff=elem->pkinfo.timeout*60*60*HZ; break;
			case 'd': elem->timediff=elem->pkinfo.timeout*60*60*24*HZ; break;
			default:
				  printk(KERN_ERR "sigknock: unknown timeout_char '%c', this can't happen\n",elem->pkinfo.timeout_char);
				  return 0;
		}
	}
	else if(elem->pkinfo.close_after_syn==0)
	{
		printk(KERN_ERR "sigknock: neither timeout nor close_after_syn supplied, can't happen\n");
		return 0;
	}
	elem->authorized=0;

	/*
	 * we want to modify the list, so grab the lock.
	 */
	spin_lock_irqsave(&info_list_lock, flags);   
	list_add(&elem->list, &info_list);
	spin_unlock_irqrestore(&info_list_lock, flags);

#ifdef ULTRADEBUG
	spin_lock_irqsave(&info_list_lock, flags);
	dump_infolist(&info_list);
	spin_unlock_irqrestore(&info_list_lock, flags);
#endif

	return 1;
}

/*******************************************************************************/

static void ipt_pk_destroy_target(void *matchinfo, unsigned int matchsize)
{
}

/*******************************************************************************/

static void ipt_pk_destroy_match(void *matchinfo, unsigned int matchsize)
{
	struct ipt_sigknock_info *info=matchinfo;
	struct info_list_elem *elem;
	unsigned long flags;
	static int count=0;
	int found=0;

	count++;

#ifdef ULTRADEBUG
	// lacks UID adaption
	/*PRINTK("%s: removing iptables entry:\n",__FUNCTION__);
	  PRINTK("   timeout: %d%c\n   uid: %s\n",
	  info->timeout, info->timeout_char, info->uid);*/
#endif

	spin_lock_irqsave(&info_list_lock, flags);
	list_for_each_entry(elem, &info_list, list)
	{
		if(infocmp(&elem->pkinfo, info)==0)
		{
			list_del(&elem->list);
			kfree(elem);
			PRINTK("** %s (%d): kfree: %p\n",__FUNCTION__, count, elem);
			found=1;
			break;
		}
	}
	spin_unlock_irqrestore(&info_list_lock, flags);

	if(found==0)
	{
		printk(KERN_ERR "sigknock: %s: couldn't find my entry in the "
				"infolist while removing the match\n",__FUNCTION__);
	}

#ifdef ULTRADEBUG
	spin_lock_irqsave(&info_list_lock, flags);
	dump_infolist(&info_list);
	spin_unlock_irqrestore(&info_list_lock, flags);
#endif
}

/*******************************************************************************/

/*
 * structure for the sigknock-match
 */
static struct ipt_match pk_match = {
	.name		= "sigknock",
	.match		= &ipt_pk_match,
	.checkentry	= &ipt_pk_checkentry_match,
	.destroy    = &ipt_pk_destroy_match,
	.me		= THIS_MODULE,
};

/*
 * structure for the SIGKNOCK-KEYPORT-target
 */
static struct ipt_target pk_target = {
	.name="SIGKNOCK-KEYPORT",
	.target=&ipt_pk_target,
	.checkentry=&ipt_pk_checkentry_target,
	.destroy=&ipt_pk_destroy_target,
	.me=THIS_MODULE,
};

static int __init init(void)
{
	int pid, res;
	unsigned int i, d;

	/* 
	 * determine how many times we have to shift jiffies so the
	 * value only changes every 20 seconds (or more)
	 */
	for(i=0, d=1;d<=20*HZ && i<32;i++)
	{
		d=d*2;
	}
	if(i==32)
	{
		printk(KERN_ERR "sigknock: unable to determine jiffies delay, exiting\n");
		goto cleanup0;
	}
	PRINTK(KERN_INFO "sigknock: jiffies delay: %d shifts (%d secs)\n",
			i, (int)(d/HZ));
	jiffies_shift_for_delay=i;

	/*
	 * register the netfilter match
	 */
	if((res=ipt_register_match(&pk_match))!=0)
	{
		printk(KERN_ERR "sigknock: error ipt_register_match: %d\n",res);
		goto cleanup0;
	}

	/*
	 * register the netfilter target
	 */
	if((res=ipt_register_target(&pk_target))!=0)
	{
		printk(KERN_ERR "sigknock: error ipt_register_target: %d\n",res);
		goto cleanup1;
	}

	/*
	 * register procfs entries
	 */
	if((proc_dir=create_proc_entry("ipt_sigknock", S_IFDIR, proc_net))==NULL)
	{
		printk(KERN_ERR "sigknock: error registering proc entry\n");
		goto cleanup2;
	}
	if((proc_challenge=create_proc_entry("challenge", 0400, proc_dir))==NULL)
	{
		printk(KERN_ERR "sigknock: error registering proc entry\n");
		goto cleanup3;
	}
	if((proc_signature=create_proc_entry("signature", 0400, proc_dir))==NULL)
	{
		printk(KERN_ERR "sigknock: error registering proc entry\n");
		goto cleanup4;
	}
	if((proc_result=create_proc_entry("result", 0200, proc_dir))==NULL)
	{
		printk(KERN_ERR "sigknock: error registering proc entry\n");
		goto cleanup5;
	}
	proc_challenge->read_proc=proc_read_challenge;
	proc_challenge->data=NULL;
	proc_signature->read_proc=proc_read_signature;
	proc_signature->data=NULL;
	proc_result->write_proc=proc_write_result;
	proc_result->data=NULL;

	/*
	 * start signature checking kernel thread
	 */
	pid=kernel_thread(signature_thread,NULL,0);
	if(pid<0)
	{
		printk(KERN_ERR "sigknock: error starting signature kernel thread: %d\n",pid);
		goto cleanup6;
	}
	else
	{
		printk(KERN_INFO "sigknock: started kernel thread with PID %d\n",pid);
		sigthread_pid=pid;
	}

	return 0;

cleanup6:
	remove_proc_entry("result", proc_dir);
cleanup5:
	remove_proc_entry("signature", proc_dir);
cleanup4:
	remove_proc_entry("challenge", proc_dir);
cleanup3:
	remove_proc_entry("ipt_sigknock", proc_net);
cleanup2:
	ipt_unregister_target(&pk_target);
cleanup1:
	ipt_unregister_match(&pk_match);
cleanup0:
	return -1;
}

/*******************************************************************************/

static void __exit fini(void)
{
	ipt_unregister_match(&pk_match);
	ipt_unregister_target(&pk_target);

	/*
	 * terminate the signature thread
	 */
	init_MUTEX_LOCKED(&sem_exit_notify);
	exit_signature_thread_command=1;

	/* 
	 * wake up the thread so it can terminate
	 */
	wake_up(&signature_wait_queue);

	/*
	 * wait for termination of the thread
	 */
	down(&sem_exit_notify);

	/*
	 * remove proc entries
	 */
	remove_proc_entry("signature", proc_dir);
	remove_proc_entry("result", proc_dir);
	remove_proc_entry("challenge", proc_dir);
	remove_proc_entry("ipt_sigknock", proc_net);
}

module_init(init);
module_exit(fini);
